@@ -1,9 +1,6 @@
 /*
  * (c) 1998-2000 Gerd Knorr
- *
- *    capture a image, compress as jpeg and upload to the webserver
- *    q
- *    using ftp the ftp utility
+ * (c) 2000-2004 Tom Gilbert
  *
  */
 
@@ -11,6 +8,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <math.h>
 #include <stdarg.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -21,6 +19,7 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <sys/select.h>
 #include <signal.h>
 #include <X11/Xlib.h>
 #include <Imlib2.h>
@@ -37,9 +36,9 @@
 #include "parseconfig.h"
 #include "pwc-ioctl.h"
 
-#define VERSION "1.3"
+#define VERSION "1.9"
 
-void log(char *fmt,
+void camlog(char *fmt,
          ...);
 
 char *ftp_host = "www";
@@ -53,7 +52,9 @@ char *temp_file = "/tmp/webcam.jpg";
 pid_t childpid = 0;
 int ftp_passive = 1;
 int ftp_do = 1;
+int ftp_upload_every = 1; /* default to upload every shot */
 int ftp_keepalive = 1;
+int ftp_delete_first = 0;
 char *scp_target = NULL;
 char *grab_device = "/dev/video0";
 char *grab_text = "";           /* strftime */
@@ -64,7 +65,7 @@ char *grab_infofile = NULL;
 char *logfile = NULL;
 int grab_width = 320;
 int grab_height = 240;
-int grab_delay = 3;
+float grab_delay = 3.0;
 int grab_quality = 75;
 int lag_reduce = 5;
 int scp_timeout = 30;
@@ -93,7 +94,13 @@ char *title_font = "arial/8";
 char *ttf_dir = "/usr/X11R6/lib/X11/fonts/TrueType";
 char *archive_ext = "jpg";
 char *grab_archive = NULL;
-int archive_shot_every = 1; /* default to archive every shot */
+int archive_subdirs = 0;    /* default to archive without    */
+                            /* subdirs                       */
+int archive_shot_every = 1;     /* default to archive every shot */
+char *archive_thumbnails_dir = NULL;
+int archive_thumbnails_create = 0; /* default is not to create archive thumbnails */
+int archive_thumbnails_width = 120;
+int archive_thumbnails_height = 90;
 char *grab_blockfile = NULL;
 char *upload_blockfile = NULL;
 char *grab_postprocess = NULL;
@@ -111,6 +118,9 @@ char *watch_interface = NULL;
 char *offline_image = NULL;
 int interface_active = 0;
 int device_palette;
+char *pwc_wb_mode = "auto";
+int pwc_wb_red = 50;
+int pwc_wb_blue = 50;
 
 int crop = 0;
 int crop_x = 0;
@@ -129,6 +139,9 @@ int v_force = 0;
 int bw_percent = 100;
 int delay_correct = 0;
 int reinit_device = 0;
+int flip_horizontal = 0;
+int flip_vertical = 0;
+int orientation = 0;
 
 int connections = 0;
 CURL *curl_handle = NULL;
@@ -153,6 +166,44 @@ void alarm_handler(int sig);
 int save_image(Imlib_Image image,
                char *file);
 
+
+struct timeval add_timevals( struct timeval t1, struct timeval t2 )
+{
+  struct timeval tr;
+  tr.tv_usec = t1.tv_usec + t2.tv_usec;
+  tr.tv_sec = t1.tv_sec + t2.tv_sec + tr.tv_usec/1000000;
+  tr.tv_usec = tr.tv_usec%1000000;
+  return tr;
+}
+
+struct timeval sub_timevals( struct timeval t1, struct timeval t2 )
+{
+  struct timeval tr;
+  tr.tv_usec = t1.tv_usec - t2.tv_usec;
+  tr.tv_sec = t1.tv_sec - t2.tv_sec + t1.tv_usec/1000000;
+  while( tr.tv_usec<0 ){
+    tr.tv_sec-=1;
+    tr.tv_usec+=1000000;
+  }
+  return tr;
+}
+
+float float_timeval( struct timeval t )
+{
+  float f = 1e-6 * t.tv_usec;
+  f += t.tv_sec;
+  return f;
+}
+
+struct timeval timeval_float( float f )
+{
+  struct timeval tr;
+  tr.tv_sec = (time_t) floor(f);
+  f -= tr.tv_sec;
+  tr.tv_usec = (suseconds_t) round(1e6*f);
+  return tr;
+}
+
 void
 close_device()
 {
@@ -174,12 +225,12 @@ try_palette(int fd,
   cam_pic.palette = pal;
   cam_pic.depth = depth;
   if (ioctl(fd, VIDIOCSPICT, &cam_pic) < 0)
-    return FALSE;
+    return 0;
   if (ioctl(fd, VIDIOCGPICT, &cam_pic) < 0)
-    return FALSE;
+    return 0;
   if (cam_pic.palette == pal)
-    return TRUE;
-  return FALSE;
+    return 1;
+  return 0;
 }
 
 int
@@ -187,15 +238,15 @@ find_palette(int fd,
              struct video_mmap *vid)
 {
   if (try_palette(fd, VIDEO_PALETTE_RGB24, 24)) {
-    log("negotiated palette RGB24\n");
+    camlog("negotiated palette RGB24\n");
     return VIDEO_PALETTE_RGB24;
   }
   if (try_palette(fd, VIDEO_PALETTE_YUV420P, 16)) {
-    log("negotiated palette YUV420P\n");
+    camlog("negotiated palette YUV420P\n");
     return VIDEO_PALETTE_YUV420P;
   }
   if (try_palette(fd, VIDEO_PALETTE_YUV420, 16)) {
-    log("negotiated palette YUV420\n");
+    camlog("negotiated palette YUV420\n");
     return VIDEO_PALETTE_YUV420;
   }
   fprintf(stderr,
@@ -238,16 +289,17 @@ grab_init()
   grab_buf.height = grab_height;
 
   ioctl(grab_fd, VIDIOCGMBUF, &vid_mbuf);
-  log("%s detected\n", grab_cap.name);
+  camlog("%s detected\n", grab_cap.name);
 
   /* special philips features */
   if (sscanf(grab_cap.name, "Philips %d webcam", &type) > 0) {
     struct video_window vwin;
     int shutter = -1;
     int gain = -1;
+    struct pwc_whitebalance wb;
 
     /* philips cam detected, maybe enable special features */
-    log("enabling pwc-specific features\n");
+    camlog("enabling pwc-specific features\n");
 
     ioctl(grab_fd, VIDIOCGWIN, &vwin);
     if (vwin.flags & PWC_FPS_MASK) {
@@ -266,6 +318,28 @@ grab_init()
       perror("trying to set gain");
     if (ioctl(grab_fd, VIDIOCPWCSSHUTTER, &shutter) < 0)
       perror("trying to set shutter");
+
+    wb.mode = PWC_WB_AUTO;
+    wb.manual_red = 50;
+    wb.manual_blue = 50;
+    if (!strcasecmp(pwc_wb_mode, "auto")) {
+      wb.mode = PWC_WB_AUTO;
+    } else if (!strcasecmp(pwc_wb_mode, "indoor")) {
+      wb.mode = PWC_WB_INDOOR;
+    } else if (!strcasecmp(pwc_wb_mode, "outdoor")) {
+      wb.mode = PWC_WB_OUTDOOR;
+    } else if (!strcasecmp(pwc_wb_mode, "fluorescent")) {
+      wb.mode = PWC_WB_FL;
+    } else if (!strcasecmp(pwc_wb_mode, "manual")) {
+      wb.mode = PWC_WB_MANUAL;
+      wb.manual_red = 65535 * ((float) pwc_wb_red / 100);
+      wb.manual_blue = 65535 * ((float) pwc_wb_blue / 100);
+    } else {
+      camlog("unknown pwc white balance mode '%s' ignored\n", pwc_wb_mode);
+    }
+
+    if (ioctl(grab_fd, VIDIOCPWCSAWB, &wb) < 0)
+      perror("trying to set pwc white balance mode");
   }
 
   /* set image source and TV norm */
@@ -285,7 +359,7 @@ grab_init()
   grab_size = vid_mbuf.size;
   grab_data =
     mmap(0, grab_size, PROT_READ | PROT_WRITE, MAP_SHARED, grab_fd, 0);
-  if ((grab_data == NULL) || (-1 == (int) grab_data)) {
+  if ((grab_data == NULL) || (MAP_FAILED == grab_data)) {
     fprintf(stderr,
             "couldn't mmap vidcam. your card doesn't support that?\n");
     exit(1);
@@ -597,7 +671,7 @@ do_postprocess(char *filename)
   if (grab_postprocess) {
     char buf[4096];
 
-    log("executing postprocessing\n");
+    camlog("executing postprocessing\n");
     snprintf(buf, sizeof(buf), "%s %s", grab_postprocess, filename);
     system(buf);
   }
@@ -606,33 +680,106 @@ do_postprocess(char *filename)
 void
 archive_jpeg(Imlib_Image im)
 {
-  char buffer[1028];
+  char buffer[PATH_MAX];
   char date[128];
+  char year[5];
+  char month[3];
+  char day[3];
   time_t t;
   struct tm *tm;
-  struct stat st;
   static int shot_counter = 0;
+  Imlib_Image thumbnail_image;
 
   shot_counter++;
 
-  if (grab_archive && archive_shot_every 
-                   && shot_counter >= archive_shot_every) {
+  if (grab_archive && archive_shot_every
+      && shot_counter >= archive_shot_every) {
     time(&t);
     tm = localtime(&t);
-    strftime(date, 127, "%Y-%m-%d_%H%M%S", tm);
 
-    do {
+    if (archive_subdirs) {
+      strftime(date, 128, "%H%M%S", tm);
+      strftime(year, 5, "%Y", tm);
+      strftime(month, 3, "%m", tm);
+      strftime(day, 3, "%d", tm);
+
+      snprintf(buffer, sizeof(buffer), "%s/%s", grab_archive, year);
+      if (access(buffer, F_OK) == -1) {
+        mkdir(buffer, 0777);
+        camlog("Created new archive subdir %s\n", buffer);
+      }
+
+      snprintf(buffer, sizeof(buffer), "%s/%s/%s", grab_archive, year,
+               month);
+      if (access(buffer, F_OK) == -1) {
+        mkdir(buffer, 0777);
+        camlog("Created new archive subdir %s\n", buffer);
+      }
+
+      snprintf(buffer, sizeof(buffer), "%s/%s/%s/%s", grab_archive, year,
+               month, day);
+      if (access(buffer, F_OK) == -1) {
+        mkdir(buffer, 0777);
+        camlog("Created new archive subdir %s\n", buffer);
+      }
+
+      snprintf(buffer, sizeof(buffer), "%s/%s/%s/%s/%s.%s", grab_archive,
+               year, month, day, date, archive_ext);
+    } else {
+      strftime(date, 127, "%Y-%m-%d_%H%M%S", tm);
       snprintf(buffer, sizeof(buffer), "%s/webcam_%s.%s", grab_archive, date,
-               archive_ext);
+              archive_ext);
     }
-    while (stat(buffer, &st) == 0);
     save_image(im, buffer);
     shot_counter = 0;
+    /* 
+     * if archive thumbnails are enabled we save them here,
+     * same filenames just in "archive_thumbnails_directory" directory
+     * files are saved in the same format as archive
+     */
+    if (archive_thumbnails_create && archive_thumbnails_width
+        && archive_thumbnails_height) {
+      if (archive_subdirs) {
+        snprintf(buffer, sizeof(buffer), "%s/%s", archive_thumbnails_dir, year);
+        if (access(buffer, F_OK) == -1) {
+          mkdir(buffer, 0777);
+          camlog("Created new archive subdir %s\n", buffer);
+        }
+
+        snprintf(buffer, sizeof(buffer), "%s/%s/%s", archive_thumbnails_dir,
+                 year, month);
+        if (access(buffer, F_OK) == -1) {
+          mkdir(buffer, 0777);
+          camlog("Created new archive subdir %s\n", buffer);
+        }
+
+        snprintf(buffer, sizeof(buffer), "%s/%s/%s/%s",
+                 archive_thumbnails_dir, year, month, day);
+        if (access(buffer, F_OK) == -1) {
+          mkdir(buffer, 0777);
+          camlog("Created new archive subdir %s\n", buffer);
+        }
+
+        snprintf(buffer, sizeof(buffer), "%s/%s/%s/%s/%s.%s",
+                 archive_thumbnails_dir, year, month, day, date, archive_ext);
+      } else {
+        snprintf(buffer, sizeof(buffer), "%s/webcam_%s.%s",
+          archive_thumbnails_dir, date, archive_ext);
+      }
+      thumbnail_image =
+        gib_imlib_create_cropped_scaled_image(im, 0, 0,
+                                              gib_imlib_image_get_width(im),
+                                              gib_imlib_image_get_height(im),
+                                              archive_thumbnails_width,
+                                              archive_thumbnails_height, 1);
+      save_image(thumbnail_image, buffer);
+      gib_imlib_free_image_and_decache(thumbnail_image);
+    }
   }
 }
 
 void
-log(char *fmt,
+camlog(char *fmt,
     ...)
 {
   va_list args;
@@ -675,52 +822,52 @@ save_image(Imlib_Image image,
   if ((err) || (!image)) {
     switch (err) {
       case IMLIB_LOAD_ERROR_FILE_DOES_NOT_EXIST:
-        log("Error saving image %s - File does not exist", file);
+        camlog("Error saving image %s - File does not exist", file);
         break;
       case IMLIB_LOAD_ERROR_FILE_IS_DIRECTORY:
-        log("Error saving image %s - Directory specified for image filename",
+        camlog("Error saving image %s - Directory specified for image filename",
             file);
         break;
       case IMLIB_LOAD_ERROR_PERMISSION_DENIED_TO_READ:
-        log("Error saving image %s - No read access to directory", file);
+        camlog("Error saving image %s - No read access to directory", file);
         break;
       case IMLIB_LOAD_ERROR_UNKNOWN:
       case IMLIB_LOAD_ERROR_NO_LOADER_FOR_FILE_FORMAT:
-        log("Error saving image %s - No Imlib2 loader for that file format",
+        camlog("Error saving image %s - No Imlib2 loader for that file format",
             file);
         break;
       case IMLIB_LOAD_ERROR_PATH_TOO_LONG:
-        log("Error saving image %s - Path specified is too long", file);
+        camlog("Error saving image %s - Path specified is too long", file);
         break;
       case IMLIB_LOAD_ERROR_PATH_COMPONENT_NON_EXISTANT:
-        log("Error saving image %s - Path component does not exist", file);
+        camlog("Error saving image %s - Path component does not exist", file);
         break;
       case IMLIB_LOAD_ERROR_PATH_COMPONENT_NOT_DIRECTORY:
-        log("Error saving image %s - Path component is not a directory",
+        camlog("Error saving image %s - Path component is not a directory",
             file);
         break;
       case IMLIB_LOAD_ERROR_PATH_POINTS_OUTSIDE_ADDRESS_SPACE:
-        log("Error saving image %s - Path points outside address space",
+        camlog("Error saving image %s - Path points outside address space",
             file);
         break;
       case IMLIB_LOAD_ERROR_TOO_MANY_SYMBOLIC_LINKS:
-        log("Error saving image %s - Too many levels of symbolic links",
+        camlog("Error saving image %s - Too many levels of symbolic links",
             file);
         break;
       case IMLIB_LOAD_ERROR_OUT_OF_MEMORY:
-        log("Error saving image %s - Out of memory", file);
+        camlog("Error saving image %s - Out of memory", file);
         break;
       case IMLIB_LOAD_ERROR_OUT_OF_FILE_DESCRIPTORS:
-        eprintf("While loading %s - Out of file descriptors", file);
+        gib_eprintf("While loading %s - Out of file descriptors", file);
         break;
       case IMLIB_LOAD_ERROR_PERMISSION_DENIED_TO_WRITE:
-        log("Error saving image %s - Cannot write to directory", file);
+        camlog("Error saving image %s - Cannot write to directory", file);
         break;
       case IMLIB_LOAD_ERROR_OUT_OF_DISK_SPACE:
-        log("Error saving image %s - Cannot write - out of disk space", file);
+        camlog("Error saving image %s - Cannot write - out of disk space", file);
         break;
       default:
-        log
+        camlog
           ("Error saving image %s - Unknown error (%d). Attempting to continue",
            file, err);
         break;
@@ -743,13 +890,10 @@ draw_overlay(Imlib_Image image)
 };
 
 void
-bw_res_change(int diff)
+bw_res_change(struct timeval diff)
 {
   int temp = 0;
 
-  if (!diff)
-    return;
-
   if (v_curr == -1) {
 
     while (temp < 5) {
@@ -763,21 +907,21 @@ bw_res_change(int diff)
               "You don't appear to be running any of the resolutions\n");
       fprintf(stderr,
               "req'd by the bandwidth limiter. It has been deactivated.\n");
-      log("method bw_percent killed, not at support'd res\n");
+      camlog("method bw_percent killed, not at support'd res\n");
     }
   }
 
-  if (diff > (grab_delay * bw_percent) / 100) {
-    log("bw_res_change Not enough bandwidth.\n");
+  if (float_timeval(diff) > (grab_delay * bw_percent) / 100) {
+    camlog("bw_res_change Not enough bandwidth.\n");
     if (v_force < -1 && v_curr > 0) {
-      log("bw_res_change Reducing image resolution.\n");
+      camlog("bw_res_change Reducing image resolution.\n");
       grab_buf.height = v_height[--v_curr];
       grab_buf.width = v_width[v_curr];
     }
     v_force--;
-  } else if (diff < (grab_delay * bw_percent) / 200) {
+  } else if (float_timeval(diff) < (grab_delay * bw_percent) / 200) {
     if (v_force > 1 && v_curr < 5) {
-      log("bw_res_change Increasing image resolution.\n");
+      camlog("bw_res_change Increasing image resolution.\n");
       grab_buf.height = v_height[++v_curr];
       grab_buf.width = v_width[v_curr];
     }
@@ -797,161 +941,172 @@ ftp_upload(char *local,
   FILE *infile;
   CURLcode ret;
   struct stat st;
-  struct curl_slist *post_commands = NULL;
-  char *passwd_string, *url_string;
+  static struct curl_slist *post_commands = NULL;
+  static char *passwd_string = NULL;
+  static char *url_string = NULL;
 
   infile = fopen(local, "r");
 
   if (!infile) {
-    fprintf(stderr, "camE: Couldn't open temp file to upload it\n");
+    camlog("camE: Couldn't open temp file to upload it\n");
     perror("ftp_upload(): ");
     return;
   }
   fstat(fileno(infile), &st);
 
-  snprintf(buf, sizeof(buf), "rnfr %s", tmp);
-  post_commands = curl_slist_append(post_commands, buf);
-  snprintf(buf, sizeof(buf), "rnto %s", remote);
-  post_commands = curl_slist_append(post_commands, buf);
+  if (!post_commands) {
+    if (ftp_delete_first) {
+      snprintf(buf, sizeof(buf), "dele %s", remote);
+      post_commands = curl_slist_append(post_commands, buf);
+    }
+    snprintf(buf, sizeof(buf), "rnfr %s", tmp);
+    post_commands = curl_slist_append(post_commands, buf);
+    snprintf(buf, sizeof(buf), "rnto %s", remote);
+    post_commands = curl_slist_append(post_commands, buf);
+  }
 
   /* init the curl session */
   if (connections < 1) {
     curl_handle = curl_easy_init();
     connections++;
+
+    if (passwd_string)
+      free(passwd_string);
+    passwd_string = gib_strjoin(":", ftp_user, ftp_pass, NULL);
+    curl_easy_setopt(curl_handle, CURLOPT_USERPWD, passwd_string);
+
+    /* set URL to save to */
+    if (url_string)
+      free(url_string);
+    url_string = gib_strjoin("/", "ftp:/", ftp_host, ftp_dir, tmp, NULL);
+    curl_easy_setopt(curl_handle, CURLOPT_URL, url_string);
+
+    /* no progress meter please */
+    curl_easy_setopt(curl_handle, CURLOPT_NOPROGRESS, 1);
+
+    /* shut up completely */
+    if (ftp_debug)
+      curl_easy_setopt(curl_handle, CURLOPT_VERBOSE, 1);
+    else
+      curl_easy_setopt(curl_handle, CURLOPT_MUTE, 1);
+
+    curl_easy_setopt(curl_handle, CURLOPT_UPLOAD, 1);
+
+    if (!ftp_passive)
+      curl_easy_setopt(curl_handle, CURLOPT_FTPPORT, ftp_interface);
+
+    curl_easy_setopt(curl_handle, CURLOPT_POSTQUOTE, post_commands);
   }
 
   curl_easy_setopt(curl_handle, CURLOPT_INFILE, infile);
   curl_easy_setopt(curl_handle, CURLOPT_INFILESIZE, st.st_size);
 
-  passwd_string = gib_strjoin(":", ftp_user, ftp_pass, NULL);
-  curl_easy_setopt(curl_handle, CURLOPT_USERPWD, passwd_string);
-
-  /* set URL to save to */
-  url_string = gib_strjoin("/", "ftp:/", ftp_host, ftp_dir, tmp, NULL);
-  curl_easy_setopt(curl_handle, CURLOPT_URL, url_string);
-
-  /* no progress meter please */
-  curl_easy_setopt(curl_handle, CURLOPT_NOPROGRESS, 1);
-
-  /* shut up completely */
-  if (ftp_debug)
-    curl_easy_setopt(curl_handle, CURLOPT_VERBOSE, 1);
-  else
-    curl_easy_setopt(curl_handle, CURLOPT_MUTE, 1);
-
-  curl_easy_setopt(curl_handle, CURLOPT_POSTQUOTE, post_commands);
-
-  curl_easy_setopt(curl_handle, CURLOPT_UPLOAD, 1);
-
-  if (!ftp_passive)
-    curl_easy_setopt(curl_handle, CURLOPT_FTPPORT, ftp_interface);
-
   /* get it! */
   ret = curl_easy_perform(curl_handle);
   /* TODO check error */
   if (ret) {
     fprintf(stderr, "\ncamE: error sending via ftp: ");
-    log("EEEE error: ");
+    camlog("camE error: ");
     switch (ret) {
       case CURLE_URL_MALFORMAT:
         fprintf(stderr, "Badly formatted ftp host or directory\n");
-        log("Badly formatted ftp host or directory\n");
+        camlog("Badly formatted ftp host or directory\n");
         break;
       case CURLE_URL_MALFORMAT_USER:
         fprintf(stderr, "Badly formatted ftp username\n");
-        log("Badly formatted ftp username\n");
+        camlog("Badly formatted ftp username\n");
         break;
       case CURLE_COULDNT_RESOLVE_PROXY:
         fprintf(stderr, "Couldn't resolve proxy\n");
-        log("Couldn't resolve proxy\n");
+        camlog("Couldn't resolve proxy\n");
         break;
       case CURLE_COULDNT_RESOLVE_HOST:
         fprintf(stderr, "Unable to resolve ftp host\n");
-        log("Unable to resolve ftp host\n");
+        camlog("Unable to resolve ftp host\n");
         break;
       case CURLE_COULDNT_CONNECT:
         fprintf(stderr, "Unable to connect to ftp host\n");
-        log("Unable to connect to ftp host\n");
+        camlog("Unable to connect to ftp host\n");
         break;
       case CURLE_FTP_WEIRD_SERVER_REPLY:
         fprintf(stderr, "Wierd server reply detected\n");
-        log("Wierd server reply detected\n");
+        camlog("Wierd server reply detected\n");
         break;
       case CURLE_FTP_ACCESS_DENIED:
         fprintf(stderr, "Access denied to ftp upload\n");
-        log("Access denied to ftp upload\n");
+        camlog("Access denied to ftp upload\n");
         break;
       case CURLE_FTP_USER_PASSWORD_INCORRECT:
         fprintf(stderr, "Incorrect password for ftp login\n");
-        log("Incorrect password for ftp login\n");
+        camlog("Incorrect password for ftp login\n");
         break;
       case CURLE_FTP_WEIRD_PASS_REPLY:
         fprintf(stderr, "Wierd password reply from server\n");
-        log("Wierd password reply from server\n");
+        camlog("Wierd password reply from server\n");
         break;
       case CURLE_FTP_WEIRD_USER_REPLY:
         fprintf(stderr, "Wierd user reply from server\n");
-        log("Wierd user reply from server\n");
+        camlog("Wierd user reply from server\n");
         break;
       case CURLE_FTP_WEIRD_PASV_REPLY:
         fprintf(stderr, "Wierd passive reply from server\n");
-        log("Wierd passive reply from server\n");
+        camlog("Wierd passive reply from server\n");
         break;
       case CURLE_FTP_CANT_GET_HOST:
         fprintf(stderr, "No route to host\n");
-        log("No route to host\n");
+        camlog("No route to host\n");
         break;
       case CURLE_FTP_COULDNT_SET_BINARY:
         fprintf(stderr, "Couldn't set binary mode\n");
-        log("Couldn't set binary mode\n");
+        camlog("Couldn't set binary mode\n");
         break;
       case CURLE_PARTIAL_FILE:
         fprintf(stderr, "Only partial file uploaded\n");
-        log("Only partial file uploaded\n");
+        camlog("Only partial file uploaded\n");
         break;
       case CURLE_FTP_WRITE_ERROR:
         fprintf(stderr, "Write error\n");
-        log("Write error\n");
+        camlog("Write error\n");
         break;
       case CURLE_FTP_QUOTE_ERROR:
         fprintf(stderr, "Misquoted ftp command - check ftp config\n");
-        log("Misquoted ftp command - check ftp config\n");
+        camlog("Misquoted ftp command - check ftp config\n");
         break;
       case CURLE_WRITE_ERROR:
         fprintf(stderr, "Write error\n");
-        log("Write error\n");
+        camlog("Write error\n");
         break;
-      case CURLE_MALFORMAT_USER:	/* the user name is illegally specified */
+      case CURLE_MALFORMAT_USER: /* the user name is illegally specified */
         fprintf(stderr, "Malformatted username\n");
-        log("Malformatted username\n");
+        camlog("Malformatted username\n");
         break;
-      case CURLE_FTP_COULDNT_STOR_FILE:	/* failed FTP upload */
+      case CURLE_FTP_COULDNT_STOR_FILE: /* failed FTP upload */
         fprintf(stderr, "Couldn't STOR the file\n");
-        log("Couldn't STOR the file\n");
+        camlog("Couldn't STOR the file\n");
         break;
-      case CURLE_READ_ERROR:	/* could open/read from file */
+      case CURLE_READ_ERROR: /* could open/read from file */
         fprintf(stderr, "Couldn't open temp file\n");
-        log("Couldn't open temp file\n");
+        camlog("Couldn't open temp file\n");
         break;
       case CURLE_OUT_OF_MEMORY:
         fprintf(stderr, "Out of memory\n");
-        log("Out of memory\n");
+        camlog("Out of memory\n");
         break;
-      case CURLE_OPERATION_TIMEOUTED:	/* the timeout time was reached */
+      case CURLE_OPERATION_TIMEOUTED: /* the timeout time was reached */
         fprintf(stderr, "Upload timed out\n");
-        log("Upload timed out\n");
+        camlog("Upload timed out\n");
         break;
-      case CURLE_FTP_PORT_FAILED:	/* FTP PORT operation failed */
+      case CURLE_FTP_PORT_FAILED: /* FTP PORT operation failed */
         fprintf(stderr, "ftp PORT failed\n");
-        log("ftp PORT failed\n");
+        camlog("ftp PORT failed\n");
         break;
       case CURLE_FILE_COULDNT_READ_FILE:
         fprintf(stderr, "Couldn't read temp file\n");
-        log("Couldn't read temp file\n");
+        camlog("Couldn't read temp file\n");
         break;
       default:
         fprintf(stderr, "unknown error, attempting to continue\n");
-        log("unknown error, attempting to continue\n");
+        camlog("unknown error, attempting to continue\n");
         break;
     }
   }
@@ -959,11 +1114,10 @@ ftp_upload(char *local,
   /* cleanup curl stuff */
   if (!ftp_keepalive) {
     curl_easy_cleanup(curl_handle);
-    curl_slist_free_all(post_commands);
     connections--;
+    curl_slist_free_all(post_commands);
+    post_commands = NULL;
   }
-  free(url_string);
-  free(passwd_string);
   fclose(infile);
 }
 
@@ -988,6 +1142,8 @@ usage(void)
   printf("usage: camE [OPTION]\n");
   printf("       -c FILE    Use config file FILE\n");
   printf("       -f         Don't fork to background\n");
+  printf("       -d         Detach after forking\n");
+  printf("       -s         Single shot\n");
   printf("       -h -v      This message\n");
   exit(0);
 }
@@ -997,20 +1153,24 @@ do_upload(char *file)
 {
   struct stat st;
   int upload_successful = 1;
+  static int shot_counter = 0;
 
-  if (ftp_do) {
+  shot_counter++;
+
+  if (ftp_do && (shot_counter >= ftp_upload_every)) {
     if ((upload_blockfile && (stat(upload_blockfile, &st) == -1))
         || !upload_blockfile) {
-      log("*** uploading via ftp\n");
+      camlog("*** uploading via ftp\n");
       ftp_upload(file, ftp_file, ftp_tmp);
-      log("shot uploaded\n");
+      camlog("shot uploaded\n");
       if (action_post_upload) {
-        log("running post upload action\n");
+        camlog("running post upload action\n");
         system(action_post_upload);
-        log("post upload action done\n");
+        camlog("post upload action done\n");
       }
     }
-  } else if (scp_target) {
+    shot_counter = 0;
+  } else if (scp_target && (shot_counter >= ftp_upload_every)) {
     char target_buf[2048];
     char cmd_buf[4096];
     char *scp_args[] = { "scp", "-BCq", NULL, NULL, NULL };
@@ -1018,7 +1178,7 @@ do_upload(char *file)
 
     if (!upload_blockfile
         || (upload_blockfile && (stat(upload_blockfile, &st) == -1))) {
-      log("uploading via scp\n");
+      camlog("uploading via scp\n");
       snprintf(target_buf, sizeof(target_buf), "%s:%s/%s", scp_target,
                ftp_dir, ftp_tmp);
       snprintf(cmd_buf, sizeof(cmd_buf), "mv %s/%s %s/%s", ftp_dir, ftp_tmp,
@@ -1031,16 +1191,17 @@ do_upload(char *file)
         ssh_args[4] = cmd_buf;
         if ((upload_successful =
              execvp_with_timeout(scp_timeout, "ssh", ssh_args))) {
-          log("shot uploaded\n");
+          camlog("shot uploaded\n");
 
           if (action_post_upload) {
-            log("running post upload action\n");
+            camlog("running post upload action\n");
             system(action_post_upload);
-            log("post upload action done\n");
+            camlog("post upload action done\n");
           }
         }
       }
     }
+    shot_counter = 0;
   }
   return upload_successful;
 }
@@ -1055,24 +1216,42 @@ main(int argc,
   int width, height, i;
   struct stat st;
   pid_t childpid;
-  time_t start_shot;
-  time_t end_shot;
+  float f;
+  struct timeval start_shot;
+  struct timeval end_shot;
+  struct timeval new_delay;
   int just_shot = 0, upload_successful = 1;
-  int new_delay;
   FILE *fp;
   int ch;
   int dont_fork = 0;
+  int single_shot = 0;
   int offline_done = 1;
   char *config_file = NULL;
+  int detach = 0;
+  int ret;
 
-  while ((ch = getopt(argc, argv, "c:fhv")) != EOF) {
+  while ((ch = getopt(argc, argv, "c:fsdhv")) != EOF) {
     switch (ch) {
       case 'c':
         config_file = strdup(optarg);
         break;
       case 'f':
+        if (detach) {
+          fprintf(stderr, "-f does not make sense with -d specified.\n");
+          exit(1);
+        }
         dont_fork = 1;
         break;
+      case 'd':
+        if (dont_fork) {
+          fprintf(stderr, "-d does not make sense with -f specified.\n");
+          exit(1);
+        }
+        detach = 1;
+        break;
+      case 's':
+        single_shot = 1;
+        break;
       case 'h':
         usage();
         break;
@@ -1086,9 +1265,13 @@ main(int argc,
     }
   }
 
-  if (!dont_fork) {
+  if (!dont_fork && !single_shot) {
     /* fork and die */
-    if ((childpid = fork()) < 0) {
+    if (detach) {
+      ret = daemon(1,0);
+      if (ret != 0)
+	fprintf(stderr, "daemon (%s)\n", strerror(errno));
+    } else if ((childpid = fork()) < 0) {
       fprintf(stderr, "fork (%s)\n", strerror(errno));
       return (2);
     } else if (childpid > 0)
@@ -1100,6 +1283,8 @@ main(int argc,
     cfg_parse_file(config_file);
   } else {
     sprintf(filename, "%s/%s", getenv("HOME"), ".camErc");
+    memset(filename, '\0', sizeof(filename));
+    snprintf(filename, sizeof(filename)-1, "%s/%s", getenv("HOME"), ".camErc");
     cfg_parse_file(filename);
   }
 
@@ -1121,12 +1306,16 @@ main(int argc,
     ftp_debug = i;
   if (-1 != (i = cfg_get_int("ftp", "do")))
     ftp_do = i;
+  if (-1 != (i = cfg_get_int("ftp", "upload_every")))
+    ftp_upload_every = i;
   if (-1 != (i = cfg_get_int("ftp", "keepalive")))
     ftp_keepalive = i;
   if (-1 != (i = cfg_get_int("ftp", "timeout")))
     ftp_timeout = i;
   if (NULL != (val = cfg_get_str("ftp", "interface")))
     ftp_interface = val;
+  if (-1 != (i = cfg_get_int("ftp", "delete_first")))
+    ftp_delete_first = i;
 
   if (NULL != (val = cfg_get_str("scp", "target")))
     scp_target = val;
@@ -1157,6 +1346,8 @@ main(int argc,
     title_text = val;
   if (NULL != (val = cfg_get_str("grab", "archive_ext")))
     archive_ext = val;
+  if (NULL != (val = cfg_get_str("grab", "archive_thumbnails_dir")))
+    archive_thumbnails_dir = val;
   if (NULL != (val = cfg_get_str("grab", "logfile")))
     logfile = val;
   if (NULL != (val = cfg_get_str("grab", "ttf_dir")))
@@ -1177,12 +1368,14 @@ main(int argc,
     watch_interface = val;
   if (NULL != (val = cfg_get_str("grab", "offline_image")))
     offline_image = val;
+  if (NULL != (val = cfg_get_str("grab", "pwc_wb_mode")))
+    pwc_wb_mode = val;
   if (-1 != (i = cfg_get_int("grab", "width")))
     grab_width = i;
   if (-1 != (i = cfg_get_int("grab", "height")))
     grab_height = i;
-  if (-1 != (i = cfg_get_int("grab", "delay")))
-    grab_delay = i;
+  if (-1 != (f = cfg_get_float("grab", "delay")))
+    grab_delay = f;
   if (-1 != (i = cfg_get_int("grab", "correct")))
     delay_correct = 1;
   if (-1 != (i = cfg_get_int("grab", "percent")))
@@ -1255,6 +1448,24 @@ main(int argc,
     scale_height = i;
   if (-1 != (i = cfg_get_int("grab", "archive_shot_every")))
     archive_shot_every = i;
+  if (-1 != (i = cfg_get_int("grab", "archive_subdirs")))
+    archive_subdirs = i;
+  if (-1 != (i = cfg_get_int("grab", "archive_thumbnails_create")))
+    archive_thumbnails_create = i;
+  if (-1 != (i = cfg_get_int("grab", "archive_thumbnails_width")))
+    archive_thumbnails_width = i;
+  if (-1 != (i = cfg_get_int("grab", "archive_thumbnails_height")))
+    archive_thumbnails_height = i;
+  if (-1 != (i = cfg_get_int("grab", "pwc_wb_red")))
+    pwc_wb_red = i;
+  if (-1 != (i = cfg_get_int("grab", "pwc_wb_blue")))
+    pwc_wb_blue = i;
+  if (-1 != (i = cfg_get_int("grab", "flip_horizontal")))
+    flip_horizontal = 1;
+  if (-1 != (i = cfg_get_int("grab", "flip_vertical")))
+    flip_vertical = 1;
+  if (-1 != (i = cfg_get_int("grab", "orientation")))
+    orientation = i;
 
   if (cam_framerate > 60)
     cam_framerate = 60;
@@ -1269,8 +1480,8 @@ main(int argc,
   }
 
   /* print config */
-  log("camE " VERSION " - (c) 1999, 2000 Gerd Knorr, Tom Gilbert\n");
-  log("grabber config: size %dx%d, input %d, norm %d, " "jpeg quality %d\n",
+  camlog("camE " VERSION " - (c) 1999, 2000 Gerd Knorr, Tom Gilbert\n");
+  camlog("grabber config: size %dx%d, input %d, norm %d, " "jpeg quality %d\n",
       grab_width, grab_height, grab_input, grab_norm, grab_quality);
 
   imlib_context_set_direction(IMLIB_TEXT_TO_RIGHT);
@@ -1307,23 +1518,25 @@ main(int argc,
   grab_init();
 
   /* go! */
-  for (;;) {
+  do {
     just_shot = 0;
     upload_successful = 1;
-    end_shot = 0;
-    start_shot = 0;
+    end_shot.tv_sec = 0;
+    end_shot.tv_usec = 0;
+    start_shot.tv_sec = 0;
+    start_shot.tv_usec = 0;
 
     if (((grab_blockfile && (stat(grab_blockfile, &st) == -1))
          || !grab_blockfile) && check_interface(watch_interface)) {
       offline_done = 0;
-      time(&start_shot);
+      gettimeofday(&start_shot, 0);
       if (action_pre_shot) {
-        log("running pre-shot action\n");
+        camlog("running pre-shot action\n");
         system(action_pre_shot);
-        log("pre-shot action done\n");
+        camlog("pre-shot action done\n");
       }
 
-      log("* taking shot\n");
+      camlog("* taking shot\n");
       /* Prevent camera lag... */
       image = grab_one(&width, &height);
       imlib_context_set_image(image);
@@ -1345,6 +1558,11 @@ main(int argc,
         gib_imlib_free_image_and_decache(image);
         image = tmp_image;
         imlib_context_set_image(image);
+	
+	/* Set new values for width and height, else the image's 
+	   text might not show up in the correct position. */
+	width = crop_width;
+	height = crop_height;
       }
 
       if (scale) {
@@ -1362,14 +1580,39 @@ main(int argc,
                                          scale_width, scale_height, 1, 0, 0);
         gib_imlib_free_image_and_decache(image);
         image = tmp_image;
+        imlib_context_set_image(image);
+
+	/* Set new values for width and height, else the image's 
+	   text might not show up in the correct position. */
+	width = scale_width;
+	height = scale_height;
       }
 
-      log("** shot taken\n");
+      camlog("** shot taken\n");
       if (action_post_shot) {
-        log("running post-shot action\n");
+        camlog("running post-shot action\n");
         system(action_post_shot);
-        log("post-shot action done\n");
+        camlog("post-shot action done\n");
       }
+
+      if (flip_horizontal) {
+        imlib_image_flip_horizontal();
+      }
+      if (flip_vertical) {
+        imlib_image_flip_vertical();
+      }
+      
+      if (orientation && orientation > 0 && orientation < 4) {
+        imlib_image_orientate(orientation);
+	/* Changing orientation flips width and height, so we must swap them. */
+	int swap_dimensions;
+	if(orientation == 1 || orientation == 3) {
+	  swap_dimensions = height;
+	  height = width;
+	  width = swap_dimensions;
+	}
+      }
+      
       if (overlay_im)
         draw_overlay(image);
       add_time_text(image, get_message(), width, height);
@@ -1379,32 +1622,37 @@ main(int argc,
       gib_imlib_free_image_and_decache(image);
       upload_successful = do_upload(temp_file);
       just_shot = 1;
-      time(&end_shot);
+      gettimeofday(&end_shot,0);
     } else if (offline_image && !offline_done
                && ((upload_blockfile && (stat(upload_blockfile, &st) == -1))
                    || !upload_blockfile) && check_interface(watch_interface)) {
       /* blockfile was just created */
-      log("uploading offline image\n");
+      camlog("uploading offline image\n");
       offline_done = do_upload(offline_image);
-      log("OFFLINE\n");
+      camlog("OFFLINE\n");
     }
-    new_delay = grab_delay;
-    if (just_shot && upload_successful) {
-      end_shot = end_shot - start_shot;
-      if (bw_percent < 100)
-        bw_res_change(end_shot);
-      if (delay_correct && end_shot) {
-        new_delay -= end_shot;
-        if (new_delay < 0)
-          new_delay = 0;
-        log("Sleeping %d secs (corrected)\n", new_delay);
-      } else {
-        log("Sleeping %d secs\n", grab_delay);
+
+    if (!single_shot) {
+      new_delay = timeval_float(grab_delay);
+      if (just_shot && upload_successful) {
+        end_shot = sub_timevals(end_shot, start_shot);
+        if (bw_percent < 100)
+          bw_res_change(end_shot);
+        if (delay_correct) {
+          new_delay = sub_timevals(new_delay, end_shot);
+          if (float_timeval(new_delay) < 0) {
+            new_delay.tv_sec = 0;
+            new_delay.tv_usec = 0;
+          }
+          camlog("Sleeping %d secs (corrected)\n", new_delay);
+        } else {
+          camlog("Sleeping %d secs\n", grab_delay);
+        }
       }
+      if (upload_successful && (float_timeval(new_delay) > 0))
+        select( 0, 0, 0, 0, &new_delay );
     }
-    if (upload_successful && (new_delay > 0))
-      sleep(new_delay);
-  }
+  } while(!single_shot);
   return 0;
 }
 
@@ -1432,7 +1680,7 @@ execvp_with_timeout(int timeout,
     alarm(0);
     childpid = 0;
     if (!WIFEXITED(status) || WEXITSTATUS(status) != 0) {
-      log("exec failed for %s\n", file);
+      camlog("exec failed for %s\n", file);
       return 0;
     }
   }
@@ -1443,7 +1691,7 @@ void
 alarm_handler(int sig)
 {
   signal(sig, SIG_IGN);
-  log("timeout reached, abandoning\n");
+  camlog("timeout reached, abandoning\n");
   if (childpid) {
     kill(childpid, SIGTERM);
   }
